/* Copyright Lycée La Colinière (Nantes),/ 18/01/2019
Auteur : Frank ENDRES <frank.endres@ac-nantes.fr>

Ce logiciel est un programme informatique d'aide à l'affectation des 
élèves dans des groupes en fonction de leur voeux de spécialité dans le 
cadre de la réforme du bac entrant en vigueur à la rentrée 2019.

Ce logiciel est régi par la licence CeCILL-C soumise au droit français et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL-C telle que diffusée par le CEA, le CNRS et l'INRIA 
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilité au code source et des droits de copie,
de modification et de redistribution accordés par cette licence, il n'est
offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
seule une responsabilité restreinte pèse sur l'auteur du programme,  le
titulaire des droits patrimoniaux et les concédants successifs.

A cet égard  l'attention de l'utilisateur est attirée sur les risques
associés au chargement,  à l'utilisation,  à la modification et/ou au
développement et à la reproduction du logiciel par l'utilisateur étant 
donné sa spécificité de logiciel libre, qui peut le rendre complexe à 
manipuler et qui le réserve donc à des développeurs et des professionnels
avertis possédant  des  connaissances  informatiques approfondies.  Les
utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
logiciel à leurs besoins dans des conditions permettant d'assurer la
sécurité de leurs systèmes et ou de leurs données et, plus généralement, 
à l'utiliser et l'exploiter dans les mêmes conditions de sécurité. 

Le fait que vous puissiez accéder à cet en-tête signifie que vous avez 
pris connaissance de la licence CeCILL-C, et que vous en avez accepté les
termes. */

//Imports from UTF-8 or ISO-8859-15, LF or CR/LF
import haxe.Resource;
import haxe.Template;
import js.Browser;
import js.html.Document;
import js.html.DragEvent;
import js.html.Element;
import js.html.File;
import js.html.FileList;
import js.html.FileReader;
import js.html.InputEvent;
import js.html.InputElement;

@:expose
class Index {
	static var specialites : Array<Specialite> = null;
	static var eleves : Array<Eleve> = null;
	static var estTerminale : Bool;
	static var promo : Promotion;

/* ==================================================================================================== */
//Id HTML & propriétés CSS
	static inline var STRUCTURE_FILEINPUT : String = "structure_fileinput";
	static inline var STRUCTURE_FILENAME : String = "structure_filename";
	static inline var STRUCTURE_REPORT : String = "structure_report";
	static inline var STRUCTURE_ERROR : String = "structure_error";
	static inline var VOEUX_FILEINPUT : String = "voeux_fileinput";
	static inline var VOEUX_FILENAME : String = "voeux_filename";
	static inline var VOEUX_ERROR : String = "voeux_error";
	static inline var DOWNLOAD : String = "download";
	static inline var SECTION_SECONDCHOIX : String = "section_secondschoix";
	static inline var SECTION_NONAFFECTES : String = "section_nonaffectes";
	static inline var SECTION_ELEVES : String = "section_eleves";
	static inline var SECTION_SPECIALITES : String = "section_specialites";
	static inline var SECTION_ALIGNEMENTS : String = "section_alignements";
	static inline var SECTION_AFFECTATIONS : String = "section_affectations";
	static inline var NONE : String = "none";
	static inline var BLOCK : String = "block";
	static inline var BR : String = "<br/>";

/* ==================================================================================================== */
//FONCTIONS UTILITAIRES
	static inline function getElementById(id : String) : Element {
		return js.Browser.document.getElementById(id);
	}

	static function getSeparator(line : String) : String {
		var separator : String = null; //handle different CSV separators
		for (tryedSeparator in [ "\t", ";", ",", ":" ]) {
			if (line.indexOf(tryedSeparator) != -1) {
				separator = tryedSeparator;
				break;
			}
		}
		return separator;
	}

	static function split(line : String, separator : String) : Array<String> {
		var parts : Array<String> = new Array();
		var begin : Int = 0;
		var inStr : Bool = false;
		var hasQuotes : Bool = false;
		if (line.charCodeAt(line.length - 1) == 13) { //windows CR+LF (unix only have LF)
			line = line.substring(0, line.length - 1);
		}
		for (i in 0...line.length) {
			var chr = line.charAt(i);
			if (chr == "\"") {
				inStr = !inStr;
				hasQuotes = true;
			} else if (!inStr && chr == separator) {
				if (hasQuotes) {
					parts.push(line.substring(begin + 1, i - 1));
					hasQuotes = false;
				} else {
					parts.push(line.substring(begin, i));
				}
				begin = i + 1;
			}
		}
		if (hasQuotes) {
			parts.push(line.substring(begin + 1, line.length - 1));
		} else {
			parts.push(line.substring(begin, line.length));
		}
		return parts;
	}

	static function decode(str : String) : String { //UTF-8->ISO-8859
		var result : String = "";
		var xc3 : Bool = false; //195
		var xc5 : Bool = false; //197
		for (i in 0...str.length) {
			if (str.charCodeAt(i) == 195) { //0xc3 -> indique un caractère accentué ou ç
				xc3 = true;
			} else if (xc3) {
				result += String.fromCharCode(str.charCodeAt(i) + 64);
				xc3 = false;
			} else if (str.charCodeAt(i) == 197) { //œ ou Œ
				xc5 = true;
			} else if (xc5) {
				result += String.fromCharCode(str.charCodeAt(i) + 42);
				xc5 = false;
			} else {
				result += str.charAt(i);
			}
		}
		return result;
	}

/* ==================================================================================================== */
//FONCTIONS EVENEMENTIELLES
	public static function filtrer(input : InputElement, id : String) {
		var text = input.value.toLowerCase();
		for (row in getElementById(id).getElementsByTagName("tr")) {
			var content = "";
			for (column in row.getElementsByTagName("td")) {
				content += column.innerText.toLowerCase() + " ";
			}
			var found : Bool = true;
			for (str in text.split(" ")) {
				if (content.indexOf(str) == -1) {
					found = false;
					break;
				}
			}
			if (found) {
				row.style.display = "";
			} else {
				row.style.display = NONE;
			}
		}
	}

	public static function handleDragOver(e : DragEvent) {
		e.stopPropagation();
		e.preventDefault();
		e.dataTransfer.dropEffect = "copy";
	}

	public static function handleDropStructure(e : DragEvent) {
		e.stopPropagation();
		e.preventDefault();
		getElementById(STRUCTURE_ERROR).innerHTML = "";
		getElementById(STRUCTURE_ERROR).style.display = NONE;
		getElementById(STRUCTURE_REPORT).style.display = NONE;
		var fichiersDeposes : FileList = e.dataTransfer.files;
		if (fichiersDeposes.length != 1) {
			getElementById(STRUCTURE_ERROR).innerHTML = "Ne déposer qu'un seul fichier (la structure des classes).";
			getElementById(STRUCTURE_ERROR).style.display = BLOCK;
		} else {
			var fichierDepose : File = fichiersDeposes.item(0);
			if (fichierDepose.name.substr(fichierDepose.name.lastIndexOf(".")).toLowerCase() != ".csv") { //on windows type is equal to empty string instead of text/csv
				getElementById(STRUCTURE_ERROR).innerHTML = "Seuls les fichiers CSV sont supportés.";
				getElementById(STRUCTURE_ERROR).style.display = BLOCK;
			} else {
				displayStructureFilename(fichierDepose);
				loadStructure(fichierDepose);
			}
		}
	}

	public static function handleInputStructure(e : InputEvent) {
		getElementById(STRUCTURE_ERROR).innerHTML = "";
		getElementById(STRUCTURE_ERROR).style.display = NONE;
		getElementById(STRUCTURE_REPORT).style.display = NONE;
		var input : InputElement = cast getElementById(STRUCTURE_FILEINPUT);
		if (input.files.length == 1) {
			displayStructureFilename(input.files.item(0));
			loadStructure(input.files.item(0));
		}
	}

	public static function handleDropVoeux(e : DragEvent) {
		e.stopPropagation();
		e.preventDefault();
		getElementById(VOEUX_ERROR).innerHTML = "";
		getElementById(VOEUX_ERROR).style.display = NONE;
		var fichiersDeposes : FileList = e.dataTransfer.files;
		if (fichiersDeposes.length != 1) {
			getElementById(VOEUX_ERROR).innerHTML = "Ne déposer qu'un seul fichier (les voeux des élèves).";
			getElementById(VOEUX_ERROR).style.display = BLOCK;
		} else {
			var fichierDepose : File = fichiersDeposes.item(0);
			if (fichierDepose.name.substr(fichierDepose.name.lastIndexOf(".")).toLowerCase() != ".csv") { //on windows type is equal to empty string instead of text/csv
				getElementById(VOEUX_ERROR).innerHTML = "Seuls les fichiers CSV sont supportés.";
				getElementById(VOEUX_ERROR).style.display = BLOCK;
			} else {
				displayVoeuxFilename(fichierDepose);
				loadVoeux(fichierDepose);
			}
		}
	}

	public static function handleInputVoeux(e : InputEvent) {
		getElementById(VOEUX_ERROR).innerHTML = "";
		getElementById(VOEUX_ERROR).style.display = NONE;
		var input : InputElement = cast getElementById(VOEUX_FILEINPUT);
		if (input.files.length == 1) {
			displayVoeuxFilename(input.files.item(0));
			loadVoeux(input.files.item(0));
		}
	}

/* ==================================================================================================== */
//GESTION DES PARAMETRES
static function displayStructureFilename(fichierStructure : File) {
		var filename : String = "&nbsp;";
		if (fichierStructure != null) {
			filename = fichierStructure.name;
		}
		getElementById(STRUCTURE_FILENAME).innerHTML = filename;
	}

	static function displayStructure() {
		var tpl = new Template(Resource.getString(STRUCTURE_REPORT));
		specialites.iterator();
		getElementById(STRUCTURE_REPORT).innerHTML = tpl.execute(specialites);
		getElementById(STRUCTURE_REPORT).style.display = BLOCK;
	}

	public static function loadStructure(fichierStructure : File) {
		getElementById(STRUCTURE_ERROR).innerHTML = "";
		getElementById(STRUCTURE_ERROR).style.display = NONE;
		var reader = new FileReader();
		reader.onloadend = function() {
			var lines : Array<String> = decode(reader.result).split("\n");
			if (lines.length > 0) {
				var separator = getSeparator(lines[0]);
				if (separator == null) {
					getElementById(STRUCTURE_ERROR).innerHTML = "Le fichier est mal formatté ; utiliser <tabulation>, ',', ';' ou ':' pour séparer les champs.";
					getElementById(STRUCTURE_ERROR).style.display = BLOCK;
				} else {
					specialites = new Array();
					for (i in 0...lines.length) {
						var specialite = obtenirSpecialite(lines[i], separator, i + 1);
						if (specialite != null) {
							specialites.push(specialite);
						}
					}
					if (specialites.length == 0) {
						getElementById(STRUCTURE_ERROR).innerHTML += "Le fichier ne contient aucune ligne valide ; il faut 4 colonnes :<br/>" +
							"&nbsp;&nbsp;&nbsp;&nbsp;Spécialité, Effectif classe, Effectif réduit (ou 0), Nombre de groupes (ou 0)";
						getElementById(STRUCTURE_ERROR).style.display = BLOCK;
						specialites = null;
					} else {
						displayStructure();
						repartition();
					}
				}
			}
		}
		reader.readAsBinaryString(fichierStructure.slice());
	}

	static function obtenirSpecialite(line : String, separator : String, idx : Int) : Specialite {
		var specialite : Specialite = null;
		if (line.length != 0) {
			var fields : Array<String> = split(line, separator);
			if (fields.length != 4) {
				getElementById(STRUCTURE_ERROR).innerHTML += 'Ligne $idx invalide (' + fields.length + " champs / 4 requis) : " + line + BR;
				getElementById(STRUCTURE_ERROR).style.display = BLOCK;
			} else {
				var effectifClasse = Std.parseInt(fields[1]);
				var effectifReduit = Std.parseInt(fields[2]);
				var maxGroupes = Std.parseInt(fields[3]);
				if (effectifClasse != null && effectifReduit != null && maxGroupes != null) {
					specialite = new Specialite(fields[0] /*nom*/, effectifClasse, effectifReduit, maxGroupes);
				} else if (idx != 1) { //on passe la première ligne si elle contient les titres des colonnes
					getElementById(STRUCTURE_ERROR).innerHTML += 'Ligne $idx invalide (le champ 2, 3 ou 4 n\'est pas un nombre) : ' + line + BR;
				}
			}
		}
		return specialite;
	}

/* ==================================================================================================== */
//GESTION DES VOEUX
	static function displayVoeuxFilename(fichierVoeux : File) {
		var filename : String = "&nbsp;";
		if (fichierVoeux != null) {
			filename = fichierVoeux.name;
		}
		getElementById(VOEUX_FILENAME).innerHTML = filename;
	}

	public static function loadVoeux(fichierVoeux : File) {
		getElementById(DOWNLOAD).style.display = NONE;
		getElementById(VOEUX_ERROR).innerHTML = "";
		getElementById(VOEUX_ERROR).style.display = NONE;
		var reader = new FileReader();
		reader.onloadend = function() {
			var lines : Array<String> = decode(reader.result).split("\n");
			if (lines.length > 0) {
				var separator = getSeparator(lines[0]);
				if (separator == null) {
					getElementById(VOEUX_ERROR).innerHTML = "Le fichier est mal formatté ; utiliser <tabulation>, ',', ';' ou ':' pour séparer les champs.";
					getElementById(VOEUX_ERROR).style.display = BLOCK;
				} else {
					estTerminale = split(lines[0], separator).length == 4;
					eleves = new Array();
					for (i in 0...lines.length) {
						var eleve = obtenirEleve(lines[i], separator, i + 1);
						if (eleve != null) {
							eleves.push(eleve);
						}
					}
					if (eleves.length == 0) {
						getElementById(VOEUX_ERROR).innerHTML += "Le fichier ne contient aucune ligne valide ; il faut au moins 4 colonnes :<br/>" +
							"&nbsp;&nbsp;&nbsp;&nbsp;Nom, Prénom, Spé1, Spé2, Spé3 (première), Spé4 (première / second choix)";
						getElementById(VOEUX_ERROR).style.display = BLOCK;
						eleves = null;
					} else {
						repartition();
					}
				}
			}
		}
		reader.readAsBinaryString(fichierVoeux.slice());
	}

	static function obtenirEleve(line : String, separator : String, idx : Int) : Eleve {
		var eleve : Eleve = null;
		if (line.length != 0) {
			var fields : Array<String> = split(line, separator);
			var nbVoeuxEleve = fields.length - 2;
			if (estTerminale && nbVoeuxEleve != 2 || !estTerminale && nbVoeuxEleve != 3 && nbVoeuxEleve != 4) {
				getElementById(VOEUX_ERROR).innerHTML +=
					'Ligne $idx invalide (' + fields.length + " champs / " + Std.string(estTerminale ? 4 : 5) + " requis) : " + line + BR;
				getElementById(VOEUX_ERROR).style.display = BLOCK;
			} else {
				var nom = fields.shift();
				var prenom = fields.shift();
				//on passe la première ligne si elle contient les titres des colonnes:
				if (!(idx == 1 && nom.toLowerCase().indexOf("nom") != -1 && (prenom.toLowerCase().indexOf("prénom") != -1
					|| prenom.toLowerCase().indexOf("prenom") != -1))) {
					eleve = new Eleve(nom, prenom, fields);
				}
			}
		}
		return eleve;
	}

/* ==================================================================================================== */
//REPARTITION DES ELEVES
	static function repartition() {
		if (eleves != null) {
			var promos : Array<Promotion> = new Array();

			var promoSansTri = new Promotion(eleves, specialites, estTerminale);
			promoSansTri.creerAlignements();
			promoSansTri.affecterEleves();

			var promoTriEleves = new Promotion(eleves, specialites, estTerminale);
			promoTriEleves.creerAlignements();
			promoTriEleves.trierEleves();
			promoTriEleves.affecterEleves();
			promos.push(promoTriEleves);

			var promoTriSpecialites = new Promotion(eleves, specialites, estTerminale);
			promoTriSpecialites.trierSpecialites();
			promoTriSpecialites.creerAlignements();
			promoTriSpecialites.trierEleves();
			promoTriSpecialites.affecterEleves();
			promos.push(promoTriSpecialites);

			promo = promoSansTri;
			for (pro in promos) {
				if (pro.nonAffectes.length < promo.nonAffectes.length ||
						(pro.nonAffectes.length == promo.nonAffectes.length &&
						pro.affectesSecondChoix.length < promo.affectesSecondChoix.length)) {
					promo = pro;
				}
			}

			controle();
			affichage();
			getElementById(DOWNLOAD).style.display = BLOCK;
		}
	}

	static function affichage() {
		var nbClasses = Math.ceil(promo.eleves.length / 35);
		Reflect.setField(promo, "nbClasses", nbClasses);
		var dotationHoraire = nbClasses * 12; //12=4x3=6x2
		Reflect.setField(promo, "dotationHoraire", dotationHoraire);
		var consoDotation : Int = 0;
		var nbGroupes : Int = 0;
		for (spe in promo.specialites) {
			consoDotation += spe.nbGroupesClasse * spe.coutEffectifClasse + spe.nbGroupesReduits * spe.coutEffectifReduit;
			nbGroupes += spe.nbGroupesClasse + spe.nbGroupesReduits;
		}
		Reflect.setField(promo, "consoDotation", consoDotation);
		Reflect.setField(promo, "nbGroupes", nbGroupes);

		var tpl = new Template(Resource.getString(SECTION_ELEVES));
		getElementById(SECTION_ELEVES).innerHTML = tpl.execute(promo);
		var tpl = new Template(Resource.getString(SECTION_SPECIALITES));
		getElementById(SECTION_SPECIALITES).innerHTML = tpl.execute(promo);
		var tpl = new Template(Resource.getString(SECTION_ALIGNEMENTS));
		getElementById(SECTION_ALIGNEMENTS).innerHTML = tpl.execute(promo);
		var tpl = new Template(Resource.getString(SECTION_AFFECTATIONS));
		getElementById(SECTION_AFFECTATIONS).innerHTML = tpl.execute(promo);
		if (promo.affectesSecondChoix.length == 0) {
			getElementById(SECTION_SECONDCHOIX).innerHTML = "";
		} else {
			var tpl = new Template(Resource.getString(SECTION_SECONDCHOIX));
			getElementById(SECTION_SECONDCHOIX).innerHTML = tpl.execute(promo);
		}
		if (promo.nonAffectes.length == 0) {
			getElementById(SECTION_NONAFFECTES).innerHTML = "";
		} else {
			var tpl = new Template(Resource.getString(SECTION_NONAFFECTES));
			getElementById(SECTION_NONAFFECTES).innerHTML = tpl.execute(promo);
		}
	}

	public static function download() {
		var a = Browser.document.createElement("a");
		a.setAttribute("href", "data:application/zip;base64," + Export.getEncodedZip(promo));
		a.setAttribute(DOWNLOAD, "export-respe-" + Date.now().toString().substring(0, 10) + ".zip");
		a.style.display = NONE;
		getElementById(DOWNLOAD).parentNode.appendChild(a);
		a.click();
		getElementById(DOWNLOAD).parentNode.removeChild(a);
	}

/* ==================================================================================================== */
//TESTS UNITAIRES
	static function controle() {
		assert(promo.getEffectifsSpecialites() == (eleves.length * (estTerminale ? 2 : 3)), "dénombrement des élèves : " +
			Std.string(promo.getEffectifsSpecialites()) + "<>" + Std.string(eleves.length * (estTerminale ? 2 : 3)));
		var nbAffectes = 0;
		for (ali in promo.alignements) {
			for (spe in ali) {
				for (grp in spe.groupes) {
					nbAffectes += grp.elevesAffectes.length;
				}
			}
		}
		assert(eleves.length == (promo.nonAffectes.length + nbAffectes / (estTerminale ? 2 : 3)), "affectation des élèves : " +
			Std.string(eleves.length) + "<>" + Std.string(promo.nonAffectes.length + nbAffectes / (estTerminale ? 2 : 3)));
	}

	static function assert(test : Bool, message : String) {
		if (!test) {
			Browser.alert("Erreur fatale : " + message);
			throw(message);
		}
	}
}
